# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from unittest import mock

from absl.testing import flagsaver
from etils import epath
import pytest
from tensorflow_datasets import testing
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import naming
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import version as version_lib

_DATA_DIR = epath.Path('/a')
_DATASET_NAME = 'my_ds'
_DATASET_DIR = _DATA_DIR / _DATASET_NAME
_VERSION = '1.0.0'


def test_default_data_dir():
  data_dir = constants.get_default_data_dir()
  assert data_dir


def _create_dataset_dir(
    data_dir: epath.Path | None = None, version: str = _VERSION
) -> epath.Path:
  dataset_dir = file_utils.get_dataset_dir(
      data_dir=data_dir,
      builder_name=_DATASET_NAME,
      config_name=None,
      version=version,
  )
  dataset_dir.mkdir(parents=True, exist_ok=True)
  return dataset_dir


def _assert_data_dir(
    given_data_dir: epath.Path | None, expected_data_dir: epath.Path
) -> None:
  data_dir, _ = file_utils.get_data_dir_and_dataset_dir(
      given_data_dir=given_data_dir,
      builder_name=_DATASET_NAME,
      config_name=None,
      version=_VERSION,
  )
  assert data_dir == expected_data_dir


@pytest.fixture(name='other_data_dir')
def mock_other_data_dir(default_data_dir: epath.Path):
  """Adds another data dir to the registered data dirs."""
  other_data_dir = default_data_dir.parent / 'other_data_dir'
  file_utils.add_data_dir(other_data_dir)
  yield other_data_dir
  file_utils.clear_registered_data_dirs()


def test_get_data_dir(default_data_dir: epath.Path):
  # No data_dir is passed
  # -> use default data_dir
  _assert_data_dir(given_data_dir=None, expected_data_dir=default_data_dir)


def test_get_data_dir_multi_dir(
    default_data_dir: epath.Path, other_data_dir: epath.Path
):
  del other_data_dir
  # No data_dir is passed
  # -> use default data_dir
  _assert_data_dir(given_data_dir=None, expected_data_dir=default_data_dir)


def test_get_data_dir_multi_dir_explicit(other_data_dir: epath.Path):
  # data_dir is passed
  # -> use data_dir
  _assert_data_dir(
      given_data_dir=other_data_dir, expected_data_dir=other_data_dir
  )


def test_get_data_dir_multi_dir_old_version_exists(
    default_data_dir: epath.Path, other_data_dir: epath.Path
):
  # Other data dir contains old versions
  _create_dataset_dir(data_dir=other_data_dir, version='0.1.0')
  _create_dataset_dir(data_dir=other_data_dir, version='0.2.0')
  _create_dataset_dir(data_dir=default_data_dir)
  # No data_dir is passed
  # -> use default data_dir
  _assert_data_dir(given_data_dir=None, expected_data_dir=default_data_dir)


def test_get_data_dir_multi_dir_version_exists(
    other_data_dir: epath.Path,
):
  # Only one data dir contains the version
  _create_dataset_dir(data_dir=other_data_dir)
  # No data_dir is passed
  # -> use existing data
  _assert_data_dir(given_data_dir=None, expected_data_dir=other_data_dir)


def test_get_data_dir_multi_dir_version_exists_explicit(
    default_data_dir: epath.Path, other_data_dir: epath.Path
):
  # Two data dirs contain the version
  _create_dataset_dir(data_dir=default_data_dir)
  _create_dataset_dir(data_dir=other_data_dir)
  # data_dir is passed
  # -> use data_dir
  _assert_data_dir(
      given_data_dir=other_data_dir, expected_data_dir=other_data_dir
  )


def test_get_data_dir_multi_dir_version_exists_raises(
    default_data_dir: epath.Path, other_data_dir: epath.Path
):
  # Two data dirs contain the version
  _create_dataset_dir(data_dir=default_data_dir)
  _create_dataset_dir(data_dir=other_data_dir)

  with pytest.raises(ValueError, match='found in more than one directory'):
    file_utils.get_data_dir_and_dataset_dir(
        given_data_dir=None,
        builder_name=_DATASET_NAME,
        config_name=None,
        version=_VERSION,
    )


def _add_dataset_info(mock_fs: testing.MockFs, dataset_dir: epath.Path) -> None:
  mock_fs.add_file(dataset_dir / constants.DATASET_INFO_FILENAME)


def _add_features(
    mock_fs: testing.MockFs, dataset_dir: epath.Path, content: str = None
) -> None:
  mock_fs.add_file(dataset_dir / constants.FEATURES_FILENAME, content=content)


def _touch(path: epath.Path) -> None:
  path.parent.mkdir(parents=True, exist_ok=True)
  path.touch()


def test_list_dataset_versions(tmpdir):
  tmpdir = epath.Path(tmpdir)
  _touch(tmpdir / '1.0.0' / constants.DATASET_INFO_FILENAME)
  _touch(tmpdir / '1.0.1' / constants.DATASET_INFO_FILENAME)
  _touch(tmpdir / '3.0.0' / constants.DATASET_INFO_FILENAME)

  # Does not have dataset_info.json, so ignored.
  _touch(tmpdir / '4.0.0' / 'other_file.json')

  # Version folder is inside a subfolder, so ignored.
  _touch(tmpdir / 'xxx' / '1.0.0' / constants.DATASET_INFO_FILENAME)

  # Subfolder name is not a valid version, so ignored.
  _touch(tmpdir / 'not_valid_version' / constants.DATASET_INFO_FILENAME)

  actual_versions = file_utils.list_dataset_versions(tmpdir)
  assert actual_versions == [
      version_lib.Version('1.0.0'),
      version_lib.Version('1.0.1'),
      version_lib.Version('3.0.0'),
  ]


def test_list_dataset_variants_with_configs(mock_fs: testing.MockFs):
  configs_and_versions = {
      'x': [_VERSION, '1.0.1'],
      'y': ['2.0.0'],
  }
  info_filenames = {
      constants.FEATURES_FILENAME,
      constants.DATASET_INFO_FILENAME,
  }
  for config, versions in configs_and_versions.items():
    for version in versions:
      for info_filename in info_filenames:
        mock_fs.add_file(_DATASET_DIR / config / version / info_filename)

  references = sorted(
      file_utils.list_dataset_variants(dataset_dir=_DATASET_DIR)
  )
  assert references == [
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          config='x',
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          config='x',
          version='1.0.1',
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          config='y',
          version='2.0.0',
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
  ]


def test_list_dataset_variants_without_configs(mock_fs: testing.MockFs):
  # Version 1.0.0 doesn't have features.json, because it was generated with an
  # old version of TFDS.
  _add_dataset_info(mock_fs, _DATASET_DIR / _VERSION)
  _add_dataset_info(mock_fs, _DATASET_DIR / '1.0.1')
  _add_features(mock_fs, _DATASET_DIR / '1.0.1')

  # List dirs including datasets generated by old TFDS versions.
  references = sorted(
      file_utils.list_dataset_variants(
          dataset_dir=_DATASET_DIR,
          include_old_tfds_version=True,
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames={constants.DATASET_INFO_FILENAME},
      ),
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          version='1.0.1',
          data_dir=_DATA_DIR,
          info_filenames={
              constants.DATASET_INFO_FILENAME,
              constants.FEATURES_FILENAME,
          },
      ),
  ]

  # List dirs excluding datasets generated by old TFDS versions.
  references = sorted(
      file_utils.list_dataset_variants(
          dataset_dir=_DATASET_DIR,
          include_old_tfds_version=False,
      )
  )
  assert references == [
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          version='1.0.1',
          data_dir=_DATA_DIR,
          info_filenames={
              constants.DATASET_INFO_FILENAME,
              constants.FEATURES_FILENAME,
          },
      )
  ]


def test_list_datasets_in_data_dir(mock_fs: testing.MockFs):
  _add_dataset_info(mock_fs, _DATA_DIR / 'ds1' / 'config1' / _VERSION)
  _add_features(mock_fs, _DATA_DIR / 'ds1' / 'config1' / _VERSION)
  _add_dataset_info(mock_fs, _DATA_DIR / 'ds1' / 'config1' / '2.0.0')
  _add_features(mock_fs, _DATA_DIR / 'ds1' / 'config1' / '2.0.0')
  _add_dataset_info(mock_fs, _DATA_DIR / 'ds1' / 'config2' / _VERSION)
  _add_features(mock_fs, _DATA_DIR / 'ds1' / 'config2' / _VERSION)

  _add_dataset_info(mock_fs, _DATA_DIR / 'ds2' / _VERSION)
  _add_features(mock_fs, _DATA_DIR / 'ds2' / _VERSION)

  info_filenames = {
      constants.DATASET_INFO_FILENAME,
      constants.FEATURES_FILENAME,
  }

  # The following are problematic and should thus be ignored.
  _add_features(mock_fs, _DATA_DIR / 'invalid-name' / _VERSION, content='x')
  _add_features(mock_fs, _DATA_DIR / 'invalid_version1' / '1.a.b', content='x')
  _add_features(
      mock_fs, _DATA_DIR / 'invalid_version2' / '1.2.3.4', content='x'
  )

  references = sorted(file_utils.list_datasets_in_data_dir(data_dir=_DATA_DIR))
  assert references == [
      naming.DatasetReference(
          dataset_name='ds1',
          config='config1',
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
      naming.DatasetReference(
          dataset_name='ds1',
          config='config1',
          version='2.0.0',
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
      naming.DatasetReference(
          dataset_name='ds1',
          config='config2',
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
      naming.DatasetReference(
          dataset_name='ds2',
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames=info_filenames,
      ),
  ]


def test_list_datasets_in_data_dir_with_namespace(mock_fs: testing.MockFs):
  namespace = 'ns'
  _add_dataset_info(mock_fs, _DATASET_DIR / 'config1' / _VERSION)
  _add_features(mock_fs, _DATASET_DIR / 'config1' / _VERSION)

  references = sorted(
      file_utils.list_datasets_in_data_dir(
          data_dir=_DATA_DIR,
          namespace=namespace,
          include_configs=True,
          include_versions=True,
      )
  )

  assert references == [
      naming.DatasetReference(
          dataset_name=_DATASET_NAME,
          namespace=namespace,
          config='config1',
          version=_VERSION,
          data_dir=_DATA_DIR,
          info_filenames={
              constants.DATASET_INFO_FILENAME,
              constants.FEATURES_FILENAME,
          },
      ),
  ]


def test_find_files_without_glob(mock_fs: testing.MockFs):
  folder = epath.Path('/')
  mock_fs.add_file(folder / 'a' / 'b' / 'x')
  mock_fs.add_file(folder / 'a' / 'c' / 'x')
  mock_fs.add_file(folder / 'b' / 'd' / 'x')
  mock_fs.add_file(folder / 'b' / 'd' / 'y')  # Should be ignored.
  mock_fs.add_file(folder / 'b' / '.config' / 'x')  # Should be ignored.
  mock_fs.add_file(folder / 'b' / 'x')
  mock_fs.add_file(folder / 'b' / 'y')  # Should be ignored.
  actual = file_utils._find_files_without_glob(
      folder, globs=['*/*', '*/*/*'], file_names=['x']
  )
  actual = [os.fspath(p) for p in actual]
  assert sorted(actual) == ['/a/b/x', '/a/c/x', '/b/d/x', '/b/x']


@pytest.mark.parametrize(
    ['filename', 'result'],
    [
        ('abc', False),
        (constants.DATASET_INFO_FILENAME, True),
        (constants.FEATURES_FILENAME, True),
        ('mnist-test.tfrecord-00000-of-00001', True),
        ('mnist-test.arrayrecord-00000-of-00001', True),
    ],
)
def test_looks_like_a_tfds_file(filename, result):
  assert file_utils._looks_like_a_tfds_file(filename) == result


@pytest.mark.parametrize(
    ['path', 'glob_result', 'expected'],
    [
        ('/a/*', ['/a/b', '/a/c'], ['/a/b', '/a/c']),
        ('/a/b', None, ['/a/b']),
        ('a/*', None, ['a/*']),
        ('/a/b@*', None, ['/a/b@*']),
    ],
)
def test_expand_glob(path, glob_result, expected):
  with mock.patch.object(epath, 'Path') as mock_epath:
    mock_epath.return_value.expanduser.return_value = path
    mock_epath.return_value.glob.return_value = glob_result
    actual = file_utils.expand_glob(path)
    if glob_result is not None:
      mock_epath.return_value.glob.assert_called_once_with(path[1:])
    else:
      mock_epath.return_value.glob.assert_not_called()
    actual = [os.fspath(p) for p in actual]
    assert actual == expected


def test_publish_data(mock_fs: testing.MockFs):
  from_data_dir = epath.Path('/tmp') / 'dummy_mnist/3.0.1'
  filename = constants.DATASET_INFO_FILENAME
  content = 'a'
  mock_fs.add_file(path=from_data_dir / filename, content=content)
  to_data_dir = epath.Path('/a/b')
  file_utils.publish_data(from_data_dir=from_data_dir, to_data_dir=to_data_dir)
  assert mock_fs.read_file(to_data_dir / filename) == content


class BulkOperationTest(testing.TestCase):

  def test_bulk_delete(self):
    tmp_dir = epath.Path(self.tmp_dir)
    file_1 = tmp_dir / 'a'
    file_2 = tmp_dir / 'b'
    file_1.touch()
    file_2.touch()
    file_utils.bulk_delete([file_1, file_2])
    self.assertFalse(file_1.exists())
    self.assertFalse(file_2.exists())

  def test_bulk_rename(self):
    tmp_dir = epath.Path(self.tmp_dir)
    orig_files = [tmp_dir / f'src{i}' for i in range(10)]
    dst_files = [tmp_dir / f'dst{i}' for i in range(10)]
    for file in orig_files:
      file.touch()
    file_utils.bulk_rename(old_paths=orig_files, new_paths=dst_files)
    for file in orig_files:
      self.assertFalse(file.exists())
    for file in dst_files:
      self.assertTrue(file.exists())

  def test_bulk_rename_with_different_number_of_files(self):
    tmp_dir = epath.Path(self.tmp_dir)
    orig_files = [tmp_dir / f'src{i}' for i in range(10)]
    dst_files = [tmp_dir / f'dst{i}' for i in range(5)]
    with self.assertRaises(ValueError):
      file_utils.bulk_rename(old_paths=orig_files, new_paths=dst_files)

  def test_bulk_rename_with_same_old_and_new_paths(self):
    tmp_dir = epath.Path(self.tmp_dir)
    orig_files = [tmp_dir / f'src{i}' for i in range(10)]
    with self.assertRaises(ValueError):
      file_utils.bulk_rename(old_paths=orig_files, new_paths=orig_files)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for optional_feature."""

from absl.testing import parameterized
import numpy as np
from tensorflow_datasets import testing
from tensorflow_datasets.core import features


class ScalarFeatureTest(
    parameterized.TestCase, testing.FeatureExpectationsTestCase
):

  @parameterized.parameters(
      (np.int64, 42, 42, 42),
      (np.int64, None, -9223372036854775808, testing.TestValue.NONE),
      (np.str_, 'foo', 'foo', 'foo'),
      (np.str_, None, '', testing.TestValue.NONE),
  )
  def test_scalar(self, dtype, value, expected, expected_np):
    self.assertFeature(
        feature=features.Scalar(
            dtype=dtype, doc='Some description', optional=True
        ),
        shape=(),
        dtype=dtype,
        tests=[
            testing.FeatureExpectationItem(
                value=value,
                expected=expected,
                expected_np=expected_np,
            ),
        ],
    )

  def test_dict(self):
    self.assertFeature(
        feature=features.FeaturesDict(
            {'a': features.Scalar(np.int32, optional=True)}
        ),
        shape={'a': ()},
        dtype={'a': np.int32},
        tests=[
            testing.FeatureExpectationItem(
                value={'a': None},
                expected={'a': -2147483648},
                expected_np={'a': None},
            ),
            # You cannot ommit the key, you do have to specify {'a': None}.
            testing.FeatureExpectationItem(
                value={},
                raise_cls=RuntimeError,
                raise_cls_np=RuntimeError,
                raise_msg="'a'",
            ),
        ],
    )

  def test_raise_error_if_feature_is_not_optional(self):
    self.assertFeature(
        feature=features.FeaturesDict({'a': np.int32}),
        shape={'a': ()},
        dtype={'a': np.int32},
        tests=[
            testing.FeatureExpectationItem(
                value={'a': None},
                raise_cls=TypeError,
                raise_cls_np=TypeError,
                raise_msg="not 'NoneType'",
            ),
        ],
    )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LVIS dataset."""

from __future__ import annotations

import collections
import json
import pathlib

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_EXTRACT_URLS = {
    'train_annotation': (
        'https://dl.fbaipublicfiles.com/LVIS/lvis_v1_train.json.zip'
    ),
    'train_images': 'http://images.cocodataset.org/zips/train2017.zip',
    'validation_annotation': (
        'https://dl.fbaipublicfiles.com/LVIS/lvis_v1_val.json.zip'
    ),
    'validation_images': 'http://images.cocodataset.org/zips/val2017.zip',
    'test_annotation': (
        'https://dl.fbaipublicfiles.com/LVIS/lvis_v1_image_info_test_dev.json.zip'
    ),
    'test_images': 'http://images.cocodataset.org/zips/test2017.zip',
}
_URLS = {
    # Minival from
    #   https://gitlab.com/YuanHuan_Pro/YOLO-World/-/blob/master/docs/data.md:
    'minival_annotation': (
        'https://huggingface.co/GLIPModel/GLIP/resolve/main/lvis_v1_minival_inserted_image_name.json'
    ),
}

# Annotations with invalid bounding boxes. Will not be used.
_INVALID_ANNOTATIONS = [
    # Train split.
    662101,
    81217,
    462924,
    227817,
    29381,
    601484,
    412185,
    504667,
    572573,
    91937,
    239022,
    181534,
    101685,
    # Validation split.
    36668,
    57541,
    33126,
    10932,
]

_NUM_CLASSES = 1203


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for lvis dataset."""

  VERSION = tfds.core.Version('1.4.0')
  RELEASE_NOTES = {
      '1.1.0': (
          'Added fields `neg_category_ids` and `not_exhaustive_category_ids`.'
      ),
      '1.2.0': 'Added class names.',
      '1.3.0': 'Added minival split.',
      '1.4.0': 'Added segmentation masks to the minival split.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    class_label = tfds.features.ClassLabel(
        names_file=tfds.core.tfds_path('datasets/lvis/classes.txt')
    )
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'image/id': np.int64,
            'neg_category_ids': tfds.features.Sequence(class_label),
            'not_exhaustive_category_ids': tfds.features.Sequence(class_label),
            'objects': tfds.features.Sequence({
                # LVIS has unique id for each annotation.
                'id': np.int64,
                'area': np.int64,
                'bbox': tfds.features.BBoxFeature(),
                'label': class_label,
                'segmentation': tfds.features.Image(shape=(None, None, 1)),
            }),
        }),
        # If there's a common (input, target) tuple from the
        # features, specify them here. They'll be used if
        # `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=None,
        homepage='https://www.lvisdataset.org/',
        description=(
            'LVIS: A dataset for large vocabulary instance'
            ' segmentation.\n\nOfficial splits: train, validation, and test.'
            ' The minival split was introduced by MDETR'
            ' (https://arxiv.org/abs/2104.12763;'
            ' https://github.com/ashkamath/mdetr/blob/main/.github/lvis.md).'
        ),
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    paths = {
        **dl_manager.download_and_extract(_EXTRACT_URLS),
        **dl_manager.download(_URLS),
    }
    image_dirs = [
        paths['train_images'] / 'train2017',
        paths['validation_images'] / 'val2017',
        paths['test_images'] / 'test2017',
    ]
    return {
        tfds.Split.TRAIN: self._generate_examples(
            image_dirs, paths['train_annotation'] / 'lvis_v1_train.json'
        ),
        tfds.Split.VALIDATION: self._generate_examples(
            image_dirs, paths['validation_annotation'] / 'lvis_v1_val.json'
        ),
        tfds.Split.TEST: self._generate_examples(
            image_dirs,
            paths['test_annotation'] / 'lvis_v1_image_info_test_dev.json',
        ),
        'minival': self._generate_examples(
            image_dirs,
            paths['minival_annotation'],
        ),
    }

  def _generate_examples(self, image_dirs, annotation_file):
    """Yields examples."""
    lvis_annotation = LvisAnnotation(annotation_file)

    def _process_example(image_info):
      # Search image dirs.
      filename = pathlib.Path(image_info['coco_url']).name
      image = _find_image_in_dirs(image_dirs, filename)
      instances = lvis_annotation.get_annotations(img_id=image_info['id'])
      instances = [x for x in instances if x['id'] not in _INVALID_ANNOTATIONS]
      neg_category_ids = image_info.get('neg_category_ids', [])
      not_exhaustive_category_ids = image_info.get(
          'not_exhaustive_category_ids', []
      )
      example = {
          'image': image,
          'image/id': image_info['id'],
          'neg_category_ids': [i - 1 for i in neg_category_ids],
          'not_exhaustive_category_ids': [
              i - 1 for i in not_exhaustive_category_ids
          ],
          'objects': [],
      }
      for inst in instances:
        example['objects'].append({
            'id': inst['id'],
            'area': inst['area'],
            'bbox': _build_bbox(image_info, *inst['bbox']),
            'label': inst['category_id'] - 1,
            'segmentation': _build_segmentation_mask(
                image_info, inst['segmentation']
            ),
        })
      return image_info['id'], example

    beam = tfds.core.lazy_imports.apache_beam
    return beam.Create(lvis_annotation.images) | beam.Map(_process_example)


def _find_image_in_dirs(image_dirs, filename):
  """Finds `filename` in one of the `image_dir` folders."""
  images = [d / filename for d in image_dirs if (d / filename).exists()]
  assert len(images) == 1, (images, image_dirs, filename)
  return images[0]


def _build_bbox(image_info, x, y, width, height):
  # build_bbox is only used within the loop so it is ok to use image_info
  return tfds.features.BBox(
      ymin=y / image_info['height'],
      xmin=x / image_info['width'],
      ymax=(y + height) / image_info['height'],
      xmax=(x + width) / image_info['width'],
  )


def _build_segmentation_mask(image_info, seg):
  cv2 = tfds.core.lazy_imports.cv2
  mask = np.zeros((image_info['height'], image_info['width']), np.uint8)
  error_msg = f'Annotation contains an invalid polygon with < 3 points: {seg}'
  assert all(len(poly) % 2 == 0 and len(poly) >= 6 for poly in seg), error_msg

  for poly in seg:
    poly = np.asarray(poly, np.int32).reshape((1, -1, 2))
    cv2.fillPoly(mask, poly, 255)
  return mask[:, :, np.newaxis]


class LvisAnnotation:
  """LVIS annotation helper class.

  The format of the annations is explained on
  https://www.lvisdataset.org/dataset.
  """

  def __init__(self, annotation_path):
    with epath.Path(annotation_path).open() as f:
      data = json.load(f)
    self._data = data

    img_id2annotations = collections.defaultdict(list)
    for a in self._data.get('annotations', []):
      img_id2annotations[a['image_id']].append(a)
    self._img_id2annotations = {
        k: list(sorted(v, key=lambda a: a['id']))
        for k, v in img_id2annotations.items()
    }

  @property
  def categories(self):
    """Return the category dicts, as sorted in the file."""
    return self._data['categories']

  @property
  def images(self):
    """Return the image dicts, as sorted in the file."""
    return self._data['images']

  def get_annotations(self, img_id):
    """Return all annotations associated with the image id string."""
    # Some images don't have any annotations. Return empty list instead.
    return self._img_id2annotations.get(img_id, [])



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import inspect

import tensorflow_datasets as tfds

_DO_NOT_CKECK_DATASETS = ['imagenet_v2']


@dataclasses.dataclass(frozen=True, kw_only=True)
class _BuilderWithVersionMismatch:
  name: str
  file: str
  version: str
  max_version_in_release_notes: str


def test_internal_datasets_have_versions_on_line_with_the_release_notes():
  builders = tfds.list_builders(with_community_datasets=False)
  assert builders
  builders_with_version_mismatch: list[_BuilderWithVersionMismatch] = []
  for builder in builders:
    builder_cls = tfds.core.registered.imported_builder_cls(builder)
    if builder_cls.name in _DO_NOT_CKECK_DATASETS:
      continue
    if not (
        hasattr(builder_cls, 'VERSION')
        and hasattr(builder_cls, 'RELEASE_NOTES')
    ):
      # This can be the case for test datasets
      continue
    if builder_cls.VERSION and builder_cls.RELEASE_NOTES:
      max_version_in_release_notes = max(
          [tfds.core.Version(version) for version in builder_cls.RELEASE_NOTES]
      )
      version = tfds.core.Version(builder_cls.VERSION)
      if version < max_version_in_release_notes:
        # This means the builder is as follow:
        # ```
        # RELEASE_NOTES = {
        #   '1.0.1': 'Bug fix.',
        #   '1.0.0': 'Initial release.',
        # }
        # VERSION = '1.0.0'  # <- Someone forgot to increment this version.
        # ```
        file = inspect.getfile(builder_cls).split('tensorflow_datasets/')[-1]
        builders_with_version_mismatch.append(
            _BuilderWithVersionMismatch(
                name=builder_cls.name,
                file=file,
                version=version,
                max_version_in_release_notes=max_version_in_release_notes,
            )
        )
  if builders_with_version_mismatch:
    error = 'The following datasets have a version mismatch:'
    for builder_cls in builders_with_version_mismatch:
      error += (
          f'\n  - Dataset {builder_cls.name} ({builder_cls.file}) has VERSION='
          f'"{builder_cls.version}" but RELEASE_NOTES contains version'
          f' "{builder_cls.max_version_in_release_notes}".'
      )
    raise ValueError(error)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convert QM9 dataset to tensorflow_datasets (tfds) format.

This module converts the QM9 dataset (Ramakrishnan et al, "Quantum chemistry
structures and properties of 134 kilo molecules", Scientific Data, 2014) to
tensorflow_datasets (tfds).

To build the dataset, run the following from directory containing this file:
$ tfds build.
"""

import dataclasses
import re
from typing import Any, Iterable

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds


pd = tfds.core.lazy_imports.pandas

_HOMEPAGE = 'https://doi.org/10.6084/m9.figshare.c.978904.v5'

_ATOMREF_URL = 'https://figshare.com/ndownloader/files/3195395'
_UNCHARACTERIZED_URL = (
    'https://springernature.figshare.com/ndownloader/files/3195404'
)
_MOLECULES_URL = 'https://springernature.figshare.com/ndownloader/files/3195389'

_SIZE = 133_885
_CHARACTERIZED_SIZE = 130_831

_MAX_ATOMS = 29
_CHARGES = {'H': 1, 'C': 6, 'N': 7, 'O': 8, 'F': 9}
_LABELS = [
    'tag',
    'index',
    'A',
    'B',
    'C',
    'mu',
    'alpha',
    'homo',
    'lumo',
    'gap',
    'r2',
    'zpve',
    'U0',
    'U',
    'H',
    'G',
    'Cv',
]
# For each of these targets, we will add a second target with an
# _atomization suffix that has the thermo term subtracted.
_ATOMIZATION_TARGETS = ['U0', 'U', 'H', 'G']


def _process_molecule(
    atomref: dict[str, Any], fname: epath.PathLike
) -> dict[str, Any]:
  """Read molecule data from file."""
  with epath.Path(fname).open() as f:
    lines = f.readlines()
  num_atoms = int(lines[0].rstrip())
  frequencies = re.split(r'\s+', lines[num_atoms + 2].rstrip())
  smiles = re.split(r'\s+', lines[num_atoms + 3].rstrip())
  inchi = re.split(r'\s+', lines[num_atoms + 4].rstrip())

  labels = pd.read_table(fname, skiprows=1, nrows=1, sep=r'\s+', names=_LABELS)

  atoms = pd.read_table(
      fname,
      skiprows=2,
      nrows=num_atoms,
      sep=r'\s+',
      names=['Z', 'x', 'y', 'z', 'Mulliken_charge'],
  )

  # Correct exponential notation (6.8*^-6 -> 6.8e-6).
  for key in ['x', 'y', 'z', 'Mulliken_charge']:
    if atoms[key].values.dtype == 'object':
      # there are unrecognized numbers.
      atoms[key].values[:] = np.array(
          [float(x.replace('*^', 'e')) for x in atoms[key].values]
      )

  charges = np.pad(
      [_CHARGES[v] for v in atoms['Z'].values], (0, _MAX_ATOMS - num_atoms)
  )
  positions = np.stack(
      [atoms['x'].values, atoms['y'].values, atoms['z'].values], axis=-1
  ).astype(np.float32)
  positions = np.pad(positions, ((0, _MAX_ATOMS - num_atoms), (0, 0)))

  mulliken_charges = atoms['Mulliken_charge'].values.astype(np.float32)
  mulliken_charges = np.pad(mulliken_charges, ((0, _MAX_ATOMS - num_atoms)))

  example = {
      'num_atoms': num_atoms,
      'charges': charges,
      'Mulliken_charges': mulliken_charges,
      'positions': positions.astype(np.float32),
      'frequencies': frequencies,
      'SMILES': smiles[0],
      'SMILES_relaxed': smiles[1],
      'InChI': inchi[0],
      'InChI_relaxed': inchi[1],
      **{k: labels[k].values[0] for k in _LABELS},
  }

  # Create atomization targets by subtracting thermochemical energy of
  # each atom.
  for k in _ATOMIZATION_TARGETS:
    v = atomref[k]
    thermo = 0
    for z in atoms['Z'].values:
      thermo += v[z]
    example[f'{k}_atomization'] = example[k] - thermo

  return example


def _get_split_ids(
    uncharacterized,
    *,
    permutation_seed: int | None,
    train_size: int,
    validation_size: int,
):
  """Get train/validation/test split ids."""
  # Original data files are  1-indexed.
  characterized_ids = np.array(
      sorted(set(range(1, _SIZE + 1)) - set(uncharacterized))
  )
  assert len(characterized_ids) == _CHARACTERIZED_SIZE

  if permutation_seed is None:
    ids = characterized_ids
  else:
    rng = np.random.RandomState(permutation_seed)
    ids = rng.permutation(characterized_ids)

  split_ids = {}
  split_ids['train'] = ids[:train_size]

  if train_size < _CHARACTERIZED_SIZE and validation_size > 0:
    split_ids['validation'] = ids[train_size : train_size + validation_size]

  if train_size + validation_size < _CHARACTERIZED_SIZE:
    split_ids['test'] = ids[train_size + validation_size :]

  return split_ids


@dataclasses.dataclass
class BuilderConfig(tfds.core.BuilderConfig):
  # Seed used for permuting (shuffling) the dataset samples before generating
  # splits. Setting this to None disables shuffling, both before and after
  # splitting.
  permutation_seed: int | None = None
  # Size of the train split. Setting this to the full dataset size
  # (_CHARACTERIZED_SIZE) will disable the validation and test splits.
  train_size: int = _CHARACTERIZED_SIZE
  # Size of the validation split. The test split will contain remaining samples,
  # i.e. those that are neither in the train nor the validation split.
  validation_size: int = 0


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for QM9 dataset. See superclass for details."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  BUILDER_CONFIGS = [
      BuilderConfig(
          name='original',
          description=(
              'QM9 does not define any splits. So this variant puts the full'
              ' QM9 dataset in the train split, in the original order'
              ' (no shuffling).'
          ),
          permutation_seed=None,
          train_size=_CHARACTERIZED_SIZE,
          validation_size=0,
      ),
      BuilderConfig(
          name='cormorant',
          description=(
              'Dataset split used by Cormorant. 100,000 train,'
              ' 17,748 validation, and 13,083 test samples.'
              ' Splitting happens after shuffling with seed 0.'
              ' Paper: https://arxiv.org/abs/1906.04015.'
              ' Split:'
              ' https://github.com/risilab/cormorant/blob/master/src/cormorant/data/prepare/qm9.py'
          ),
          permutation_seed=0,
          train_size=100_000,
          validation_size=17_748,
      ),
      BuilderConfig(
          name='dimenet',
          description=(
              'Dataset split used by DimeNet. 110,000 train, 10,000 validation,'
              ' and 10,831 test samples.'
              ' Splitting happens after shuffling with seed 42.'
              ' Paper: https://arxiv.org/abs/2003.03123.'
              ' Split:'
              ' https://github.com/gasteigerjo/dimenet/blob/master/dimenet/training/data_provider.py'
          ),
          permutation_seed=42,
          train_size=110_000,
          validation_size=10_000,
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        disable_shuffling=self.builder_config.permutation_seed is None,
        features=tfds.features.FeaturesDict({
            'num_atoms': tfds.features.Tensor(shape=(), dtype=np.int64),
            'charges': tfds.features.Tensor(shape=(29,), dtype=np.int64),
            'Mulliken_charges': tfds.features.Tensor(
                shape=(29,), dtype=np.float32
            ),
            'positions': tfds.features.Tensor(shape=(29, 3), dtype=np.float32),
            'index': tfds.features.Tensor(shape=(), dtype=np.int64),
            'A': tfds.features.Tensor(shape=(), dtype=np.float32),
            'B': tfds.features.Tensor(shape=(), dtype=np.float32),
            'C': tfds.features.Tensor(shape=(), dtype=np.float32),
            'mu': tfds.features.Tensor(shape=(), dtype=np.float32),
            'alpha': tfds.features.Tensor(shape=(), dtype=np.float32),
            'homo': tfds.features.Tensor(shape=(), dtype=np.float32),
            'lumo': tfds.features.Tensor(shape=(), dtype=np.float32),
            'gap': tfds.features.Tensor(shape=(), dtype=np.float32),
            'r2': tfds.features.Tensor(shape=(), dtype=np.float32),
            'zpve': tfds.features.Tensor(shape=(), dtype=np.float32),
            'U0': tfds.features.Tensor(shape=(), dtype=np.float32),
            'U': tfds.features.Tensor(shape=(), dtype=np.float32),
            'H': tfds.features.Tensor(shape=(), dtype=np.float32),
            'G': tfds.features.Tensor(shape=(), dtype=np.float32),
            'Cv': tfds.features.Tensor(shape=(), dtype=np.float32),
            'U0_atomization': tfds.features.Tensor(shape=(), dtype=np.float32),
            'U_atomization': tfds.features.Tensor(shape=(), dtype=np.float32),
            'H_atomization': tfds.features.Tensor(shape=(), dtype=np.float32),
            'G_atomization': tfds.features.Tensor(shape=(), dtype=np.float32),
            'tag': tfds.features.Tensor(shape=(), dtype=np.str_),
            'SMILES': tfds.features.Tensor(shape=(), dtype=np.str_),
            'SMILES_relaxed': tfds.features.Tensor(shape=(), dtype=np.str_),
            'InChI': tfds.features.Tensor(shape=(), dtype=np.str_),
            'InChI_relaxed': tfds.features.Tensor(shape=(), dtype=np.str_),
            'frequencies': tfds.features.Tensor(
                shape=(None,), dtype=np.float32
            ),
        }),
        # These are returned if `as_supervised=True` in `builder.as_dataset`.
        supervised_keys=None,
        homepage=_HOMEPAGE,
    )

  def _split_generators(
      self, dl_manager: tfds.download.DownloadManager
  ) -> dict[str, Any]:
    """Returns SplitGenerators. See superclass method for details."""
    atomref = pd.read_table(
        dl_manager.download({'atomref': _ATOMREF_URL})['atomref'],
        skiprows=5,
        index_col='Z',
        skipfooter=1,
        sep=r'\s+',
        names=['Z', 'zpve', 'U0', 'U', 'H', 'G', 'Cv'],
    ).to_dict()

    uncharacterized = pd.read_table(
        dl_manager.download({'uncharacterized': _UNCHARACTERIZED_URL})[
            'uncharacterized'
        ],
        skiprows=9,
        skipfooter=1,
        sep=r'\s+',
        usecols=[0],
        names=['index'],
    ).values[:, 0]

    molecules_dir = dl_manager.download_and_extract(
        {'dsgdb9nsd': _MOLECULES_URL}
    )['dsgdb9nsd']

    split_ids = _get_split_ids(
        uncharacterized,
        permutation_seed=self.builder_config.permutation_seed,
        train_size=self.builder_config.train_size,
        validation_size=self.builder_config.validation_size,
    )

    return {
        split: self._generate_examples(ids, atomref, molecules_dir)
        for split, ids in split_ids.items()
    }

  def _generate_examples(
      self,
      ids: np.ndarray,
      atomref: dict[str, Any],
      molecules_dir: epath.Path,
  ) -> Iterable[tuple[int, dict[str, Any]]]:
    """Dataset generator. See superclass method for details."""

    for i in ids:
      entry = _process_molecule(
          atomref, molecules_dir / f'dsgdb9nsd_{i:06d}.xyz'
      )
      yield int(i), entry



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for huggingface_dataset_builder."""

from __future__ import annotations

from typing import Type, TypeVar

import immutabledict
import numpy as np
from tensorflow_datasets.core import features as feature_lib
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.utils import conversion_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import datasets as hf_datasets
from tensorflow_datasets.core.utils.lazy_imports_utils import huggingface_hub
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


_HF_DTYPE_TO_NP_DTYPE = immutabledict.immutabledict({
    'bool': np.bool_,
    'float': np.float32,
    'double': np.float64,
    'large_string': np.object_,
    'utf8': np.object_,
    'string': np.object_,
    'binary': np.bytes_,
})
_IMAGE_ENCODING_FORMAT = 'png'
_DEFAULT_IMG = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDATx\x9cc```\x00\x00\x00\x04\x00\x01\xf6\x178U\x00\x00\x00\x00IEND\xaeB`\x82'
_StrOrNone = TypeVar('_StrOrNone', str, None)


def _convert_to_np_dtype(hf_dtype: str) -> Type[np.generic]:
  """Returns the `np.dtype` scalar feature.

  Args:
    hf_dtype: Huggingface dtype.

  Raises:
    TypeError: If couldn't recognize Huggingface dtype.
  """
  if np_dtype := _HF_DTYPE_TO_NP_DTYPE.get(hf_dtype):
    return np_dtype
  elif hasattr(np, hf_dtype):
    return getattr(np, hf_dtype)
  if hf_dtype.startswith('timestamp'):
    # Timestamps are converted to seconds since UNIX epoch.
    return np.int64
  elif hasattr(tf.dtypes, hf_dtype):
    return getattr(tf.dtypes, hf_dtype)
  else:
    raise TypeError(
        f'Unrecognized type {hf_dtype}. Please open an issue if you think '
        'this is a bug.'
    )


def convert_hf_features(hf_features) -> feature_lib.FeatureConnector:
  """Converts Huggingface feature spec to a TFDS compatible feature spec.

  Args:
    hf_features: Huggingface feature spec.

  Returns:
    The TFDS compatible feature spec.

  Raises:
    ValueError: If the given Huggingface features is a list with length > 1.
    TypeError: If couldn't recognize the given feature spec.
  """
  match hf_features:
    case hf_datasets.Features() | dict():
      return feature_lib.FeaturesDict({
          name: convert_hf_features(hf_inner_feature)
          for name, hf_inner_feature in hf_features.items()
      })
    case hf_datasets.Sequence():
      return feature_lib.Sequence(
          feature=convert_hf_features(hf_features.feature)
      )
    case list():
      if len(hf_features) != 1:
        raise ValueError(f'List {hf_features} should have a length of 1.')
      return feature_lib.Sequence(feature=convert_hf_features(hf_features[0]))
    case hf_datasets.Value():
      return feature_lib.Scalar(dtype=_convert_to_np_dtype(hf_features.dtype))
    case hf_datasets.ClassLabel():
      if hf_features.names:
        return feature_lib.ClassLabel(
            names=[str(name) for name in hf_features.names]
        )
      if hf_features.names_file:
        return feature_lib.ClassLabel(names_file=hf_features.names_file)
      if hf_features.num_classes:
        return feature_lib.ClassLabel(num_classes=hf_features.num_classes)
    case hf_datasets.Translation():
      return feature_lib.Translation(languages=hf_features.languages)
    case hf_datasets.TranslationVariableLanguages():
      return feature_lib.TranslationVariableLanguages(
          languages=hf_features.languages
      )
    case hf_datasets.Image():
      return feature_lib.Image(encoding_format=_IMAGE_ENCODING_FORMAT)
    case hf_datasets.Audio():
      return feature_lib.Audio(
          sample_rate=hf_features.sampling_rate,
          dtype=np.int32,
      )
    case hf_datasets.Video():
      return feature_lib.Video()

  raise TypeError(f'Type {type(hf_features)} is not supported.')


def to_huggingface_name(tfds_dataset_name: str) -> str:
  """Converts TFDS dataset name to a Huggingface compatible dataset name.

  As TFDS doesn't support case-sensitive names, we list all HF datasets and pick
  the dataset that has a case-insensitive match.

  Args:
    tfds_dataset_name: TFDS dataset name.

  Returns:
    The Huggingface compatible dataset name.

  Raises:
    DatasetNotFoundError: If the TFDS dataset name doesn't correspond to any
    existing Huggingface dataset.
  """
  for hf_dataset_info in huggingface_hub.list_datasets():
    if (
        conversion_utils.to_tfds_name(hf_dataset_info.id)
        == tfds_dataset_name.lower()
    ):
      return hf_dataset_info.id
  raise registered.DatasetNotFoundError(
      f'"{tfds_dataset_name}" is not listed in Huggingface datasets.'
  )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""segment_anything dataset."""

import csv
import json
from typing import Dict, List

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_HOMEPAGE_URL = 'https://ai.facebook.com/datasets/segment-anything-downloads'
_LOCAL_LINKS_FILE_NAME = 'segment_anything_links.txt'


def _parse_json(path: epath.Path):
  with path.open() as file:
    return json.load(file)


def _normalize_bbox(
    original_coords: List[int], img_height: int, img_width: int
):
  """Converts from COCO to TFDS format."""
  if len(original_coords) != 4:
    raise ValueError(f'Malformed BBox: {original_coords}')
  # COCO format: [x_min, y_min, width, height] in absolute values.
  x_min, y_min, bbox_width, bbox_height = original_coords
  # TFDS format: [y_min, x_min, y_max, x_max] in normalized values.
  x_max = x_min + bbox_width
  y_max = y_min + bbox_height
  return np.array([
      y_min / img_height,
      x_min / img_width,
      y_max / img_height,
      x_max / img_width,
  ])


def _parse_annotations(original_annotations, img_height: int, img_width: int):
  """Parses annotations to the TFDS format."""
  annotations = []
  for annotation in original_annotations:
    annotation['bbox'] = _normalize_bbox(
        annotation['bbox'], img_height, img_width
    )
    annotation['crop_box'] = _normalize_bbox(
        annotation['crop_box'], img_height, img_width
    )
    annotations.append(annotation)
  return annotations


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for segment_anything dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }
  MANUAL_DOWNLOAD_INSTRUCTIONS = (
      f'Download the links file from {_HOMEPAGE_URL}. `manual_dir` should'
      f' contain the links file saved as {_LOCAL_LINKS_FILE_NAME}.'
  )
  # The dataset is ~1000 * 10.5GB = ~10.5TB big, so we try to increase
  # parallelism to download. The default is 50.
  MAX_SIMULTANEOUS_DOWNLOADS = 100

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    annotation_feature = tfds.features.FeaturesDict({
        'area': tfds.features.Scalar(
            dtype=np.uint64, doc='The area in pixels of the mask.'
        ),
        'bbox': tfds.features.BBoxFeature(
            doc='The box around the mask, in TFDS format.'
        ),
        'crop_box': tfds.features.BBoxFeature(
            doc=(
                'The crop of the image used to generate the mask,'
                ' in TFDS format.'
            )
        ),
        'id': tfds.features.Scalar(
            dtype=np.uint64, doc='Identifier for the annotation.'
        ),
        'point_coords': tfds.features.Tensor(
            shape=(1, 2),
            dtype=np.float64,
            doc=(
                'The point coordinates input to the model to generate the mask.'
            ),
        ),
        'predicted_iou': tfds.features.Scalar(
            dtype=np.float64,
            doc="The model's own prediction of the mask's quality.",
        ),
        'segmentation': tfds.features.FeaturesDict(
            {
                'counts': np.str_,
                'size': tfds.features.Tensor(shape=(2,), dtype=np.uint64),
            },
            doc=(
                'Encoded segmentation mask in COCO RLE format (dict'
                ' with keys `size` and `counts`).'
            ),
        ),
        'stability_score': tfds.features.Scalar(
            dtype=np.float64, doc="A measure of the mask's quality."
        ),
    })
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.FeaturesDict({
                'content': tfds.features.Image(doc='Content of the image.'),
                'file_name': np.str_,
                'height': np.uint64,
                'image_id': np.uint64,
                'width': np.uint64,
            }),
            'annotations': tfds.features.Sequence(annotation_feature),
        }),
        homepage=_HOMEPAGE_URL,
        license=f'Custom (see: {_HOMEPAGE_URL})',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path_to_links = dl_manager.manual_dir / _LOCAL_LINKS_FILE_NAME
    dict_of_urls = {}
    with path_to_links.open() as file:
      reader = csv.reader(file, delimiter='\t')
      for i, (name, url) in enumerate(reader):
        # Skip the headers.
        if i == 0:
          continue
        dict_of_urls[name] = url
    paths = dl_manager.download_and_extract(dict_of_urls)
    return {
        'train': self._generate_examples(paths),
    }

  def _generate_examples(self, paths: Dict[str, epath.Path]):
    """Yields examples."""

    def _extract_examples(path: epath.Path) -> List[epath.Path]:
      return list(path.glob('*.jpg'))

    def _process_example(image_path: epath.Path):
      json_content = _parse_json(image_path.parent / f'{image_path.stem}.json')
      image = json_content['image']
      image['content'] = image_path
      height, width = image['height'], image['width']
      assert height != 0 and width != 0, f'The image {image_path} is malformed.'
      annotations = _parse_annotations(
          json_content['annotations'], height, width
      )
      return image['image_id'], {
          'image': image,
          'annotations': annotations,
      }

    beam = tfds.core.lazy_imports.apache_beam
    return (
        beam.Create(paths.values())
        | beam.FlatMap(_extract_examples)
        | beam.Map(_process_example)
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default values for some parameters of the API when no values are passed."""

# IMPORTANT: when changing values here, update docstrings.

import os
from typing import Final

# Directory in which datasets are declared within TFDS sources.
DATASETS_TFDS_SRC_DIR = 'datasets'

# Github base URL
SRC_BASE_URL = 'https://github.com/tensorflow/datasets/tree/master/'

# Directory where to store processed datasets.
# If modifying this, should also update `scripts/cli/build.py` `--data_dir`


def get_default_data_dir() -> str:
  """Returns the TFDS default data directory."""
  return os.environ.get(
      'TFDS_DATA_DIR',
      os.path.join(os.path.expanduser('~'), 'tensorflow_datasets'),
  )


DATA_DIR: Final[str] = get_default_data_dir()

# Prefix of files / directories which aren't finished downloading / extracting.
INCOMPLETE_PREFIX = 'incomplete.'

# Note: GCS constants are defined in `core/utils/gcs_utils.py`

# Name of the file to output the features information.
FEATURES_FILENAME = 'features.json'

# Name of the file to output the DatasetInfo protobuf object.
DATASET_INFO_FILENAME = 'dataset_info.json'
LICENSE_FILENAME = 'LICENSE'
METADATA_FILENAME = 'metadata.json'
CHECKSUMS_FILENAME = 'checksums.tsv'

# Filepath for mapping between TFDS datasets and PapersWithCode entries.
PWC_FILENAME = 'tfds_to_pwc_links.json'
PWC_LINKS_PATH = f'scripts/documentation/{PWC_FILENAME}'

# Retry parameters. Delays are in seconds.
TFDS_RETRY_TRIES = int(os.environ.get('TFDS_RETRY_TRIES', 3))
TFDS_RETRY_INITIAL_DELAY = int(os.environ.get('TFDS_RETRY_INITIAL_DELAY', 1))
# How much to multiply the delay by for each subsequent try
TFDS_RETRY_DELAY_MULTIPLIER = int(
    os.environ.get('TFDS_RETRY_DELAY_MULTIPLIER', 2)
)
# Random noise to add to the delay (random pick between 0 and noise).
TFDS_RETRY_NOISE = float(os.environ.get('TFDS_RETRY_NOISE', 0.5))
# If the error message contains any of these substrings, retry.
TFDS_RETRY_MSG_SUBSTRINGS = os.environ.get(
    'TFDS_RETRY_MSG_SUBSTRINGS',
    (
        'deadline_exceeded,'
        '408 Request Timeout,'
        '429 Too Many Requests,'
        '500 Internal Server Error,'
        '502 Bad Gateway,'
        '503 Service Unavailable,'
        '504 Gateway Timeout,'
        '509 Bandwidth Limit Exceeded,'
        '599 Gateway Error'
    ),
).split(',')



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable, Iterator
import contextlib
import dataclasses
import functools
import multiprocessing
import os
import typing
from unittest import mock

from etils import epath
import pytest
import tensorflow_datasets as tfds
from tensorflow_datasets.core import download
from tensorflow_datasets.core import utils
from tensorflow_datasets.scripts.cli import build as build_lib
from tensorflow_datasets.scripts.cli import main

type_utils = tfds.core.utils.type_utils

_DUMMY_DATASET_PATH = tfds.core.tfds_path() / 'testing/dummy_dataset'


class DummyDatasetNoGenerate(tfds.testing.DummyDataset):

  def _generate_examples(self):
    if True:  # pylint: disable=using-constant-test
      raise NotImplementedError('Should not be called')
    yield

  @utils.classproperty
  @classmethod
  def url_infos(cls) -> dict[str, download.checksums.UrlInfo] | None:
    return {
        'http://data.org/file1.zip': download.checksums.UrlInfo(
            size=42,
            checksum='d45899d9a6a0e48afb250aac7ee3dc50e73e263687f15761d754515cd8284e0a',
            filename='file1.zip',
        ),
    }


@contextlib.contextmanager
def mock_cwd(path: epath.PathLike) -> Iterator[None]:
  """Mock the current directory."""
  path = epath.Path(path)
  assert path.exists() and path.is_dir()  # Check given path is valid cwd dir
  with mock.patch.object(os, 'getcwd', return_value=os.fspath(path)):
    yield


def _build(
    cmd_flags: str, data_dir: epath.Path, mock_download_and_prepare: bool = True
) -> list[str]:
  """Executes `tfds build {cmd_flags}` and returns the list of generated ds."""
  # Execute the command
  cmd_flags = f'--data_dir={data_dir} {cmd_flags}'
  args = main._parse_flags(f'tfds build {cmd_flags}'.split())

  original_dl_and_prepare = tfds.core.DatasetBuilder.download_and_prepare

  # Unfortunately, `mock.Mock` removes `self` from `call_args`, so we have
  # to patch the function to record the generated_ds manually.
  # See:
  # https://stackoverflow.com/questions/64792295/how-to-get-self-instance-in-mock-mock-call-args
  queue = multiprocessing.Queue()

  def _download_and_prepare(self, *args, **kwargs):
    # Remove version from generated name (as only last version can be generated)
    full_name = '/'.join(self.info.full_name.split('/')[:-1])
    queue.put(full_name)
    if mock_download_and_prepare:
      return
    else:
      return original_dl_and_prepare(self, *args, **kwargs)

  with mock.patch.object(
      tfds.core.DatasetBuilder, 'download_and_prepare', _download_and_prepare
  ):
    main.main(args)
    queue.put(None)

  generated_ds_names = []
  while full_name := queue.get():
    generated_ds_names.append(full_name)

  return generated_ds_names


@pytest.fixture(name='build')
def mock_build(
    default_data_dir: epath.Path,
) -> Callable[[str, bool], list[str]]:
  """Returns a function to execute `tfds build`."""
  return functools.partial(_build, data_dir=default_data_dir)


def test_build_single(build):
  assert build('mnist') == ['mnist']
  assert build('mnist:3.0.1') == ['mnist']
  # Keyword arguments also possible
  assert build('--datasets mnist') == ['mnist']

  with pytest.raises(tfds.core.registered.DatasetNotFoundError):
    build('unknown_dataset')

  with pytest.raises(AssertionError, match='cannot be loaded at version 1.0.0'):
    build('mnist:1.0.0')  # Can only built the last version

  with pytest.raises(ValueError, match='not have config'):
    build('mnist --config_idx 0')


def test_build_multiple(build):
  # Multiple datasets can be built in a single call
  assert build('mnist imagenet2012 cifar10') == [
      'mnist',
      'imagenet2012',
      'cifar10',
  ]
  # Keyword arguments also possible
  assert build('mnist --datasets imagenet2012 cifar10') == [
      'mnist',
      'imagenet2012',
      'cifar10',
  ]


@pytest.mark.parametrize('num_processes', range(1, 4))
def test_build_parallel(build, num_processes):
  # Order is not guaranteed
  assert set(build(f'trivia_qa --num-processes={num_processes}')) == set([
      'trivia_qa/rc',
      'trivia_qa/rc.nocontext',
      'trivia_qa/unfiltered',
      'trivia_qa/unfiltered.nocontext',
  ])


def test_build_dataset_configs(build):
  # By default, all configs are build
  assert build('trivia_qa') == [
      'trivia_qa/rc',
      'trivia_qa/rc.nocontext',
      'trivia_qa/unfiltered',
      'trivia_qa/unfiltered.nocontext',
  ]

  # If config is set, only the defined config is generated

  # --config_idx
  assert build('trivia_qa --config_idx=0') == ['trivia_qa/rc']

  # --config
  assert build('trivia_qa --config unfiltered.nocontext') == [
      'trivia_qa/unfiltered.nocontext',
  ]

  # --config Json
  config_json = '{"name":"my_config","description":"abcd"}'
  assert build(f'imdb_reviews --config {config_json}') == [
      'imdb_reviews/my_config',
  ]

  # name/config
  assert build('trivia_qa/unfiltered.nocontext') == [
      'trivia_qa/unfiltered.nocontext'
  ]

  with pytest.raises(ValueError, match='Config should only be defined once'):
    build('trivia_qa/unfiltered.nocontext --config_idx=0')

  with pytest.raises(ValueError, match='greater than number of configs'):
    build('trivia_qa --config_idx 100')


def test_exclude_datasets(build):
  # Exclude all datasets except 2
  all_ds = [b for b in tfds.list_builders() if b not in ('mnist', 'cifar10')]
  all_ds_str = ','.join(all_ds)

  assert build(f'--exclude_datasets {all_ds_str}') == [
      'cifar10',
      'mnist',
  ]

  with pytest.raises(ValueError, match="--exclude_datasets can't be used"):
    build('mnist --exclude_datasets cifar10')


def test_build_overwrite(build, default_data_dir: epath.Path):
  data_dir = default_data_dir / 'mnist/3.0.1'
  data_dir.mkdir(parents=True)
  metadata_path = tfds.core.tfds_path(
      'testing/test_data/dataset_info/mnist/3.0.1'
  )

  for f in metadata_path.iterdir():  # Copy metadata files.
    data_dir.joinpath(f.name).write_text(f.read_text())

  # By default, will skip generation if the data already exists
  assert build('mnist') == ['mnist']  # Called, but no-op
  assert data_dir.exists()

  assert build('mnist --overwrite') == ['mnist']
  assert not data_dir.exists()  # Previous data-dir has been removed


def test_max_examples_per_split_0(build, default_data_dir: epath.Path):
  assert build(
      'dummy_dataset_no_generate --max_examples_per_split 0',
      mock_download_and_prepare=False,
  ) == ['dummy_dataset_no_generate']

  builder_path = default_data_dir / 'dummy_dataset_no_generate/1.0.0'
  # Dataset has been generated
  assert builder_path.exists()
  # tf-records files have not been generated
  assert sorted(p.name for p in builder_path.iterdir()) == [
      'dataset_info.json',
      'features.json',
  ]


def test_build_files(build):
  # Make sure DummyDataset isn't registered by default
  with pytest.raises(tfds.core.registered.DatasetNotFoundError):
    build('dummy_dataset')

  with pytest.raises(FileNotFoundError, match='Could not find .* script'):
    build('')

  # cd .../datasets/dummy_dataset && tfds build
  with mock_cwd(_DUMMY_DATASET_PATH):
    assert build('') == ['dummy_dataset']

  # cd .../datasets/dummy_dataset && tfds build dummy_dataset.py
  with mock_cwd(_DUMMY_DATASET_PATH):
    assert build('dummy_dataset.py') == ['dummy_dataset']

  # cd .../datasets/ && tfds build dummy_dataset
  with mock_cwd(_DUMMY_DATASET_PATH.parent):
    assert build('dummy_dataset') == ['dummy_dataset']

  # cd .../datasets/ && tfds build dummy_dataset --imports=xxx
  # --imports is passed. so do not load dataset from file
  with mock_cwd(_DUMMY_DATASET_PATH.parent):
    with pytest.raises(tfds.core.registered.DatasetNotFoundError):
      assert build('dummy_dataset --imports=os')

  # cd .../datasets/ && tfds build dummy_dataset/dummy_dataset
  with mock_cwd(_DUMMY_DATASET_PATH.parent):
    assert build('dummy_dataset/dummy_dataset') == ['dummy_dataset']


# Somehow only with tf-nightly, `dummy_dataset` is already imported by
# community/load_test.py (with `skip_registration()`). Thus import here have
# no-effects.
@pytest.mark.skip(reason='Conflict with `load_test.py`')
def test_build_import(build):
  # DummyDataset isn't registered by default
  with pytest.raises(tfds.core.registered.DatasetNotFoundError):
    build('dummy_dataset')

  # --imports register the dataset
  ds_module = 'tensorflow_datasets.testing.dummy_dataset.dummy_dataset'
  assert build(f'dummy_dataset --imports {ds_module}') == ['dummy_dataset']


def test_download_only(build):
  with mock.patch.object(
      download.DownloadManager, 'download', autospec=True
  ) as mock_download:
    assert not build('dummy_dataset_no_generate --download_only')
    mock_download.assert_called_with(
        mock.ANY, {'file0': 'http://data.org/file1.zip'}
    )


@pytest.mark.parametrize(
    'args,download_config_kwargs',
    [
        ('--manual_dir=/a/b', {'manual_dir': epath.Path('/a/b')}),
        (
            '--manual_dir=/a/b --add_name_to_manual_dir',
            {'manual_dir': epath.Path('/a/b/x')},
        ),
        ('--extract_dir=/a/b', {'extract_dir': epath.Path('/a/b')}),
        ('--max_examples_per_split=42', {'max_examples_per_split': 42}),
        ('--register_checksums', {'register_checksums': True}),
        ('--force_checksums_validation', {'force_checksums_validation': True}),
        ('--max_shard_size_mb=128', {'max_shard_size': 128 << 20}),
        (
            '--download_config={"max_shard_size":1234}',
            {'max_shard_size': 1234},
        ),
    ],
)
def test_make_download_config(args: str, download_config_kwargs):
  args = main._parse_flags(f'tfds build x {args}'.split())
  cmd_args = typing.cast(build_lib.Args, args.command)
  actual = build_lib._make_download_config(cmd_args, dataset_name='x')
  # Ignore the beam runner
  actual = actual.replace(beam_runner=None)
  expected = tfds.download.DownloadConfig(**download_config_kwargs)
  assert dataclasses.asdict(actual) == dataclasses.asdict(expected)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CLEVR dataset."""

import collections
import json
import os

from etils import epath
import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_BASE_URL = "https://cs.stanford.edu/people/jcjohns/clevr/"
_DOWNLOAD_URL = "https://dl.fbaipublicfiles.com/clevr/CLEVR_v1.0.zip"


class Builder(tfds.core.GeneratorBasedBuilder):
  """CLEVR dataset."""

  VERSION = tfds.core.Version("3.1.0")
  SUPPORTED_VERSIONS = [
      tfds.core.Version("3.0.0"),
  ]
  RELEASE_NOTES = {
      "3.1.0": "Add question/answer text.",
  }

  def _info(self):
    features = {
        "image": tfds.features.Image(),
        "file_name": tfds.features.Text(),
        "objects": tfds.features.Sequence({
            "color": tfds.features.ClassLabel(
                names=[
                    "gray",
                    "blue",
                    "brown",
                    "yellow",
                    "red",
                    "green",
                    "purple",
                    "cyan",
                ]
            ),
            "material": tfds.features.ClassLabel(names=["rubber", "metal"]),
            "shape": tfds.features.ClassLabel(
                names=["cube", "sphere", "cylinder"]
            ),
            "size": tfds.features.ClassLabel(names=["small", "large"]),
            "rotation": tfds.features.Tensor(shape=(), dtype=np.float32),
            "3d_coords": tfds.features.Tensor(shape=(3,), dtype=np.float32),
            "pixel_coords": tfds.features.Tensor(shape=(3,), dtype=np.float32),
        }),
    }
    if self.version > "3.0.0":
      features["question_answer"] = tfds.features.Sequence({
          "question": tfds.features.Text(),
          "answer": tfds.features.Text(),
      })
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(features),
        homepage=_BASE_URL,
    )

  def _split_generators(self, dl_manager):
    """Returns splits."""
    path = dl_manager.download_and_extract(_DOWNLOAD_URL)

    images_path_dir = os.path.join(path, "CLEVR_v1.0/images")
    questions_path_dir = os.path.join(path, "CLEVR_v1.0/questions")
    scenes_path_dir = os.path.join(path, "CLEVR_v1.0/scenes")

    splits = []
    for split_name in ["train", "val", "test"]:
      name_map = {
          "train": tfds.Split.TRAIN,
          "val": tfds.Split.VALIDATION,
          "test": tfds.Split.TEST,
      }
      splits.append(
          tfds.core.SplitGenerator(
              name=name_map[split_name],
              gen_kwargs={
                  "images_dir_path": os.path.join(images_path_dir, split_name),
                  "question_file": os.path.join(
                      questions_path_dir,
                      "CLEVR_{}_questions.json".format(split_name),
                  ),
                  "scenes_description_file": os.path.join(
                      scenes_path_dir, "CLEVR_{}_scenes.json".format(split_name)
                  ),
              },
          )
      )

    return splits

  def _generate_examples(
      self, images_dir_path, question_file, scenes_description_file
  ):
    image_paths = sorted(
        [
            os.path.join(images_dir_path, filename)
            for filename in tf.io.gfile.listdir(images_dir_path)
        ]
    )

    with epath.Path(question_file).open() as f:
      questions_json = json.load(f)
    questions = collections.defaultdict(list)
    for q in questions_json["questions"]:
      questions[q["image_filename"]].append({
          "question": q["question"],
          "answer": q.get("answer", ""),  # Test set do not have answer.
      })

    if tf.io.gfile.exists(scenes_description_file):
      with epath.Path(scenes_description_file).open() as f:
        scenes_json = json.load(f)
    else:
      # if annotation file does not exist, we create empty annotations
      scenes_json = {"scenes": [{"objects": []}] * len(image_paths)}

    attrs = [
        "color",
        "material",
        "shape",
        "size",
        "rotation",
        "pixel_coords",
        "3d_coords",
    ]
    for image_path, scene in zip(image_paths, scenes_json["scenes"]):
      objects = scene["objects"]
      fname = os.path.basename(image_path)
      record = {
          "image": image_path,
          "file_name": fname,
          "question_answer": questions[fname],
          "objects": [{attr: obj[attr] for attr in attrs} for obj in objects],  # pylint: disable=g-complex-comprehension
      }
      yield fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""irc_disentanglement dataset."""

import collections
import os
from typing import List

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DOWNLOAD_URL = (
    "https://github.com/jkkummerfeld/irc-disentanglement/zipball/fd379e9"
)
_DOWNLOAD_ARCHIVE_SUBDIR = os.path.join(
    "jkkummerfeld-irc-disentanglement-fd379e9", "data"
)

_IRC_DAY_KEY = "day"
_MESSAGE_ID = "id"
_MESSAGE_TEXT = "text"
_MESSAGE_TIMESTAMP = "timestamp"
_MESSAGE_PARENTS_IDS = "parents"


def _get_day_to_paths(data_dir):
  """Prepares paths to files with raw chat messages and replies annotations.

  Args:
    data_dir: directory containing files with data. directory can be

  Returns:
    day_to_paths: dict formatted date -> dict with paths
      day_to_paths[day_str]["text"] - path to file with raw chat messages
      day_to_paths[day_str]["annot"] - path to file with replies annotations.
  """
  day_to_paths = collections.defaultdict(dict)
  for filename in tf.io.gfile.listdir(data_dir):
    filepath = os.path.join(data_dir, filename)
    day_str = filename[: len("YYYY-MM-DD")]  # e.g. 2004-12-25.train-c.raw.txt

    if "raw" in filename:
      day_to_paths[day_str]["text"] = filepath
    if "annotation" in filename:
      day_to_paths[day_str]["annot"] = filepath

  return day_to_paths


def _read_texts_file(path):
  with tf.io.gfile.GFile(path, "r") as f:
    return [line.strip() for line in f]


def _read_annot_file(path):
  """Reads file with replies annotation."""
  with tf.io.gfile.GFile(path, "r") as f:
    return [(int(first), int(second)) for first, second, _ in map(str.split, f)]


def _parse_out_timestamps(raw_texts, day_str):
  """Parsing timestamps from IRC chat messages.

  Similar logic is implemented here.
  https://github.com/jkkummerfeld/irc-disentanglement/blob/master/src/disentangle.py#L174

  Args:
    raw_texts: list of raw chat messages.
    day_str: formatted date string.

  Returns:
    texts: list of texts without timestamps.
    timestamps: list of formatted timestamps
  """
  prev_hours = 0
  timestamps, texts = [], []
  for raw_text in raw_texts:
    if raw_text.startswith("["):  # Regular messsages e.g. "[04:13]<xxx>: Hi!"
      hours = int(raw_text[1:3])
      mins = int(raw_text[4:6])

      # 12h format -> 24h format
      if hours < prev_hours:  # All messages belong to the same day and are
        hours += 12  # chronologically ordered, but AM/PM info is absent
      prev_hours = hours

      timestamps.append("{}_{:02}_{:02}".format(day_str, hours, mins))
      raw_text = raw_text[7:]
    else:  # System messages e.g. "=== xxx has joned #ubuntu"
      timestamps.append("")

    texts.append(raw_text)

  return texts, timestamps


def _get_msg_id(day, line_num):
  return "{}_{:05}".format(day, line_num)


def _prepare_examples(texts_file_path, annot_file_path, day_str):
  """Prepares examples for 1 day."""
  # Read raw data
  raw_texts = _read_texts_file(texts_file_path)
  annotations = _read_annot_file(annot_file_path)

  # Construct replies graph
  idx_to_parents = {idx: [] for idx in range(len(raw_texts))}
  for parent_msg_idx, msg_idx in annotations:
    idx_to_parents[msg_idx].append(parent_msg_idx)

  texts, timestamps = _parse_out_timestamps(raw_texts, day_str)

  for line_idx, parents in idx_to_parents.items():
    parents_ids = [_get_msg_id(day_str, parent) for parent in parents]
    yield {
        _MESSAGE_ID: _get_msg_id(day_str, line_idx),
        _MESSAGE_TEXT: texts[line_idx],
        _MESSAGE_TIMESTAMP: timestamps[line_idx],
        _MESSAGE_PARENTS_IDS: parents_ids,
    }


class Builder(tfds.core.GeneratorBasedBuilder):
  """IRC Disentanglement dataset."""

  VERSION = tfds.core.Version("2.0.0")

  def _info(self) -> tfds.core.DatasetInfo:
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                _IRC_DAY_KEY: tfds.features.Sequence(
                    tfds.features.FeaturesDict({
                        _MESSAGE_ID: tfds.features.Text(),
                        _MESSAGE_TEXT: tfds.features.Text(),
                        _MESSAGE_TIMESTAMP: tfds.features.Text(),
                        _MESSAGE_PARENTS_IDS: tfds.features.Sequence(
                            tfds.features.Text()
                        ),
                    })
                )
            }
        ),
        homepage="https://jkk.name/irc-disentanglement",
    )

  def _split_generators(
      self, dl_manager: tfds.download.DownloadManager
  ) -> List[tfds.core.SplitGenerator]:
    """Returns SplitGenerators."""
    base_dir = dl_manager.download_and_extract(_DOWNLOAD_URL)
    data_dir = os.path.join(base_dir, _DOWNLOAD_ARCHIVE_SUBDIR)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "day_to_paths": _get_day_to_paths(
                    os.path.join(data_dir, "train")
                )
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "day_to_paths": _get_day_to_paths(os.path.join(data_dir, "dev"))
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "day_to_paths": _get_day_to_paths(
                    os.path.join(data_dir, "test")
                )
            },
        ),
    ]

  def _generate_examples(self, day_to_paths):
    """Yields examples."""
    for day, paths in day_to_paths.items():
      yield day, {
          _IRC_DAY_KEY: list(
              _prepare_examples(paths["text"], paths["annot"], day)
          )
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Blood Cell Count and Detection Dataset."""

import collections
import os
import xml.etree.ElementTree as ET

from etils import epath
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_HOMEPAGE_URL = "https://github.com/Shenggan/BCCD_Dataset"
_DOWNLOAD_URL = "https://github.com/Shenggan/BCCD_Dataset/archive/v1.0.zip"
_CLASS_LABELS = [
    "RBC",
    "WBC",
    "Platelets",
]


class Builder(tfds.core.GeneratorBasedBuilder):
  """Blood Cell Count and Detection Dataset."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(
                shape=(480, 640, 3), encoding_format="jpeg"
            ),
            "image/filename": tfds.features.Text(),
            # Multiple bounding box per image
            "objects": tfds.features.Sequence({
                "label": tfds.features.ClassLabel(names=_CLASS_LABELS),
                "bbox": tfds.features.BBoxFeature(),
            }),
        }),
        homepage=_HOMEPAGE_URL,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    train_list = []
    test_list = []
    val_list = []

    dl_path = dl_manager.download_and_extract(_DOWNLOAD_URL)
    # Name of the extracted folder is 'BCCD_Dataset-1.0'
    extracted_dir_path = os.path.join(dl_path, "BCCD_Dataset-1.0")
    splits_dir_path = os.path.join(
        extracted_dir_path, "BCCD", "ImageSets", "Main"
    )

    for root, _, filename in tf.io.gfile.walk(splits_dir_path):
      for fname in filename:
        full_file_name = os.path.join(root, fname)
        with epath.Path(full_file_name).open() as f:
          for line in f:
            if fname == "train.txt":
              train_list.append(line)
            elif fname == "test.txt":
              test_list.append(line)
            elif fname == "val.txt":
              val_list.append(line)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "file_names": train_list,
                "extracted_dir_path": extracted_dir_path,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "file_names": val_list,
                "extracted_dir_path": extracted_dir_path,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "file_names": test_list,
                "extracted_dir_path": extracted_dir_path,
            },
        ),
    ]

  def _generate_examples(self, file_names, extracted_dir_path):
    """Yeilds Examples.

    Args:
      file_names: The path of the file that contains the splits
      extracted_dir_path: The path of the extracted folder

    Yields:
      Generator yielding the next examples
    """

    bbox_attrib = ["xmin", "xmax", "ymin", "ymax", "width", "height"]
    xml_list = collections.defaultdict(str)

    images_dir_path = os.path.join(extracted_dir_path, "BCCD", "JPEGImages")

    annotations_dir_path = os.path.join(
        extracted_dir_path, "BCCD", "Annotations"
    )

    def get_image_file_path(filename):
      """Returns image path."""
      return os.path.join(images_dir_path, "{}.jpg".format(filename.strip()))

    def get_annotations_file_path(filename):
      """Returns annotations file path."""
      return os.path.join(
          annotations_dir_path, "{}.xml".format(filename.strip())
      )

    # BBox attributes in range of 0.0 to 1.0
    def normalize_bbox(bbox_side, image_side):
      return min(bbox_side / image_side, 1.0)

    def build_box(attributes, n):
      return tfds.features.BBox(
          ymin=normalize_bbox(attributes["ymin"][n], attributes["height"][0]),
          xmin=normalize_bbox(attributes["xmin"][n], attributes["width"][0]),
          ymax=normalize_bbox(attributes["ymax"][n], attributes["height"][0]),
          xmax=normalize_bbox(attributes["xmax"][n], attributes["width"][0]),
      )

    def get_label(attributes, n):
      return attributes["name"][n]

    for fname in file_names:
      annotation_file_path = get_annotations_file_path(fname)
      with epath.Path(annotation_file_path).open() as f:
        xml_list[fname] = ET.parse(f)
      attributes = collections.defaultdict(list)
      for element in xml_list[fname].iter():
        # Extract necessary Bbox attributes from XML file
        # "Name" tag contains the label
        if element.tag.strip() == "name":
          attributes[element.tag.strip()].append(element.text.strip())
        elif element.tag.strip() in bbox_attrib:
          attributes[element.tag.strip()].append(float(element.text.strip()))

      key = fname
      example = {
          "image": get_image_file_path(fname),
          "image/filename": fname,
          "objects": [
              {  # pylint: disable=g-complex-comprehension
                  "label": get_label(attributes, n),
                  "bbox": build_box(attributes, n),
              }
              for n in range(len(attributes["name"]))
          ],
      }
      yield key, example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Amazon Customer Reviews Dataset --- US REVIEWS DATASET."""

import collections
import csv

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DATA_OPTIONS_V1_00 = [
    "Wireless",
    "Watches",
    "Video_Games",
    "Video_DVD",
    "Video",
    "Toys",
    "Tools",
    "Sports",
    "Software",
    "Shoes",
    "Pet_Products",
    "Personal_Care_Appliances",
    "PC",
    "Outdoors",
    "Office_Products",
    "Musical_Instruments",
    "Music",
    "Mobile_Electronics",
    "Mobile_Apps",
    "Major_Appliances",
    "Luggage",
    "Lawn_and_Garden",
    "Kitchen",
    "Jewelry",
    "Home_Improvement",
    "Home_Entertainment",
    "Home",
    "Health_Personal_Care",
    "Grocery",
    "Gift_Card",
    "Furniture",
    "Electronics",
    "Digital_Video_Games",
    "Digital_Video_Download",
    "Digital_Software",
    "Digital_Music_Purchase",
    "Digital_Ebook_Purchase",
    "Camera",
    "Books",
    "Beauty",
    "Baby",
    "Automotive",
    "Apparel",
]

_DATA_OPTIONS_V1_01 = ["Digital_Ebook_Purchase", "Books"]

_DATA_OPTIONS_V1_02 = ["Books"]

_DATA_OPTIONS = []

for entry in _DATA_OPTIONS_V1_00:
  _DATA_OPTIONS.append(entry + "_v1_00")

for entry in _DATA_OPTIONS_V1_01:
  _DATA_OPTIONS.append(entry + "_v1_01")

for entry in _DATA_OPTIONS_V1_02:
  _DATA_OPTIONS.append(entry + "_v1_02")

_DL_URLS = {
    name: f"https://s3.amazonaws.com/amazon-reviews-pds/tsv/amazon_reviews_us_{name}.tsv.gz"
    for name in _DATA_OPTIONS
}


class AmazonUSReviewsConfig(tfds.core.BuilderConfig):
  """BuilderConfig for AmazonUSReviews."""

  def __init__(self, *, data=None, **kwargs):
    """Constructs a AmazonUSReviewsConfig.

    Args:
      data: `str`, one of `_DATA_OPTIONS`.
      **kwargs: keyword arguments forwarded to super.
    """
    if data not in _DATA_OPTIONS:
      raise ValueError("data must be one of %s" % _DATA_OPTIONS)

    super(AmazonUSReviewsConfig, self).__init__(**kwargs)
    self.data = data


class Builder(tfds.core.GeneratorBasedBuilder):
  """AmazonUSReviews dataset."""

  BUILDER_CONFIGS = [
      AmazonUSReviewsConfig(  # pylint: disable=g-complex-comprehension
          name=config_name,
          description="A dataset consisting of reviews of Amazon "
          + config_name
          + " products in US marketplace. Each product has its own version as"
          " specified with it.",
          version="0.1.0",
          data=config_name,
      )
      for config_name in _DATA_OPTIONS
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                "data": collections.OrderedDict([
                    ("marketplace", np.str_),
                    ("customer_id", np.str_),
                    ("review_id", np.str_),
                    ("product_id", np.str_),
                    ("product_parent", np.str_),
                    ("product_title", np.str_),
                    ("product_category", np.str_),
                    ("star_rating", np.int32),
                    ("helpful_votes", np.int32),
                    ("total_votes", np.int32),
                    ("vine", tfds.features.ClassLabel(names=["Y", "N"])),
                    (
                        "verified_purchase",
                        tfds.features.ClassLabel(names=["Y", "N"]),
                    ),
                    ("review_headline", np.str_),
                    ("review_body", np.str_),
                    ("review_date", np.str_),
                ])
            }
        ),
        supervised_keys=None,
        homepage="https://s3.amazonaws.com/amazon-reviews-pds/readme.html",
    )

  def _split_generators(self, dl_manager):
    url = _DL_URLS[self.builder_config.name]
    path = dl_manager.download_and_extract(url)

    # There is no predefined train/val/test split for this dataset.
    return [
        tfds.core.SplitGenerator(
            name="train",
            gen_kwargs={
                "file_path": path,
            },
        ),
    ]

  def _generate_examples(self, file_path):
    """Generate features given the directory path.

    Args:
      file_path: path where the tsv file is stored

    Yields:
      The features.
    """

    with epath.Path(file_path).open() as tsvfile:
      # Need to disable quoting - as dataset contains invalid double quotes.
      reader = csv.DictReader(
          tsvfile, dialect="excel-tab", quoting=csv.QUOTE_NONE
      )
      for i, row in enumerate(reader):
        yield i, {
            "data": row,
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A dataset builder proxying data files from GCS."""

from __future__ import annotations

import abc
import os
from typing import Any

from tensorflow_datasets.core import beam_utils
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import dataset_utils
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.utils import read_config as read_config_lib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
import tree



class DatasetImporterBuilder(
    tfds.core.GeneratorBasedBuilder, skip_registration=True
):
  """DatasetBuilder for RLDS datasets which proxies data files from GCS.

  RLDS datasets are already converted to TFDS format. This
  builder just effectively downloads them from GCS.
  """

  VERSION = tfds.core.Version('0.1.0')
  RELEASE_NOTES = {
      '0.1.0': 'Initial release.',
  }
  _GCS_BUCKET = 'gs://gresearch/robotics/'
  KEYS_TO_STRIP = [
      'episode_id',
      'legacy_datasets',
      'project_name',
      'publish_timestamp',
      'session_pb',
      'ssot_session_key',
  ]


  @abc.abstractmethod
  def get_description(self):
    pass

  @abc.abstractmethod
  def get_citation(self):
    pass

  @abc.abstractmethod
  def get_homepage(self):
    pass

  @abc.abstractmethod
  def get_relative_dataset_location(self):
    pass

  def get_dataset_location(self):
    return os.path.join(
        str(self._GCS_BUCKET), self.get_relative_dataset_location()
    )

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    description = self.get_description()
    features = self.get_ds_builder().info.features

    tmp = dict(features)

    for key in self.KEYS_TO_STRIP:
      if key in tmp:
        del tmp[key]
    features = tfds.features.FeaturesDict(tmp)

    return tfds.core.DatasetInfo(
        builder=self,
        description=description,
        features=features,
        supervised_keys=None,
        homepage=self.get_homepage(),
        citation=self.get_citation(),
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    ds_builder = self.get_ds_builder()

    splits = {}
    for split, split_info in ds_builder.info.splits.items():
      splits[split] = self._generate_examples(ds_builder, split_info)
    return splits

  def _generate_examples(
      self,
      builder: dataset_builder.DatasetBuilder,
      split_info: splits_lib.SplitInfo,
  ):
    beam = tfds.core.lazy_imports.apache_beam
    split = split_info.name
    read_config = read_config_lib.ReadConfig(add_tfds_id=True)

    decode_fn = builder.info.features['steps'].feature.decode_example

    def converter_fn(example):
      # Decode the RLDS Episode and transform it to numpy.
      example_out = dict(example)
      example_out['steps'] = tf.data.Dataset.from_tensor_slices(
          example_out['steps']
      ).map(decode_fn)
      steps = list(iter(example_out['steps'].take(-1)))
      example_out['steps'] = steps

      example_out = dataset_utils.as_numpy(example_out)

      example_id = example_out['tfds_id'].decode('utf-8')
      del example_out['tfds_id']
      for key in self.KEYS_TO_STRIP:
        if key in example_out:
          del example_out[key]

      yield example_id, example_out

    return f'read_tfds_dataset@{split}' >> beam_utils.ReadFromTFDS(
        builder=builder,
        split=split,
        read_config=read_config,
        workers_per_shard=1,
        decoders={'steps': tfds.decode.SkipDecoding()},
    ) | f'convert_to_numpy@{split}' >> beam.FlatMap(converter_fn)

  def get_ds_builder(self):
    ds_location = self.get_dataset_location()
    ds_builder = tfds.builder_from_directory(ds_location)
    return ds_builder


class TFDSDatasetImporterBuilder(
    tfds.core.GeneratorBasedBuilder, skip_registration=True
):
  """DatasetBuilder for TFDS datasets which proxies data files from GCS.

  This builder just effectively downloads already converted datasets from GCS.
  """

  VERSION = tfds.core.Version('0.1.0')
  RELEASE_NOTES = {
      '0.1.0': 'Initial release.',
  }
  _GCS_BUCKET = 'gs://gresearch/robotics/'


  @abc.abstractmethod
  def get_description(self):
    pass

  @abc.abstractmethod
  def get_citation(self):
    pass

  @abc.abstractmethod
  def get_homepage(self):
    pass

  @abc.abstractmethod
  def get_relative_dataset_location(self):
    pass

  def get_dataset_location(self):
    return os.path.join(
        str(self._GCS_BUCKET), self.get_relative_dataset_location()
    )

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    description = self.get_description()
    features = self.get_ds_builder().info.features

    return tfds.core.DatasetInfo(
        builder=self,
        description=description,
        features=features,
        supervised_keys=None,
        homepage=self.get_homepage(),
        citation=self.get_citation(),
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    ds_builder = self.get_ds_builder()
    splits = {}
    for split, split_info in ds_builder.info.splits.items():
      splits[split] = self._generate_examples(ds_builder, split_info)
    return splits

  def _generate_examples(
      self,
      builder: dataset_builder.DatasetBuilder,
      split_info: splits_lib.SplitInfo,
  ):
    beam = tfds.core.lazy_imports.apache_beam
    split = split_info.name
    read_config = read_config_lib.ReadConfig(add_tfds_id=True)

    def converter_fn(example):
      example_out = tree.map_structure(to_np, example)
      example_id = example_out['tfds_id'].decode('utf-8')
      del example_out['tfds_id']

      yield example_id, example_out

    return f'read_tfds_dataset@{split}' >> beam_utils.ReadFromTFDS(
        builder=builder,
        split=split,
        read_config=read_config,
        workers_per_shard=1,
    ) | f'convert_to_numpy@{split}' >> beam.FlatMap(converter_fn)

  def get_ds_builder(self):
    ds_location = self.get_dataset_location()
    ds_builder = tfds.builder_from_directory(ds_location)
    return ds_builder


def to_np(tensor):
  """Convert tensor to numpy."""
  if isinstance(tensor, tf.Tensor) or isinstance(tensor, tf.RaggedTensor):
    return tensor.numpy()
  return tensor



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reddit dataset using tldr as summaries."""

from __future__ import annotations

import json
import os

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_URL = "https://zenodo.org/record/1043504/files/corpus-webis-tldr-17.zip?download=1"

_DOCUMENT = "content"
_SUMMARY = "summary"
_ADDITIONAL_FEATURES = [
    "author",
    "body",
    "normalizedBody",
    "subreddit",
    "subreddit_id",
    "id",
]


class Builder(tfds.core.GeneratorBasedBuilder):
  """Reddit Dataset."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {k: np.str_ for k in _ADDITIONAL_FEATURES + [_DOCUMENT, _SUMMARY]}
        ),
        supervised_keys=(_DOCUMENT, _SUMMARY),
        homepage="https://github.com/webis-de/webis-tldr-17-corpus",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_path = dl_manager.download_and_extract(_URL)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "path": os.path.join(dl_path, "corpus-webis-tldr-17.json")
            },
        )
    ]

  def _generate_examples(self, path=None):
    """Yields examples."""
    with tf.io.gfile.GFile(path, "rb") as f:
      for i, line in enumerate(f):
        # possible keys are:
        #   author: string (nullable = true)
        #   body: string (nullable = true)
        #   normalizedBody: string (nullable = true)
        #   content: string (nullable = true)
        #   content_len: long (nullable = true)
        #   summary: string (nullable = true)
        #   summary_len: long (nullable = true)
        #   id: string (nullable = true)
        #   subreddit: string (nullable = true)
        #   subreddit_id: string (nullable = true)
        #   title: string (nullable = true)
        d = json.loads(line)
        if _SUMMARY in d and _DOCUMENT in d:
          yield i, {
              k: d.get(k, "")
              for k in _ADDITIONAL_FEATURES + [_DOCUMENT, _SUMMARY]
          }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""CLI for Tensorflow Datasets.

TFDS CLI to help creates and build datasets (e.g. `tfds new my_dataset`,
`tfds build`,...)

See: https://www.tensorflow.org/datasets/cli
"""

import dataclasses
import logging as python_logging

from absl import app
from absl import flags
from absl import logging
import simple_parsing

import tensorflow_datasets.public_api as tfds

# Import commands
from tensorflow_datasets.scripts.cli import build
from tensorflow_datasets.scripts.cli import cli_utils
from tensorflow_datasets.scripts.cli import convert_format
from tensorflow_datasets.scripts.cli import croissant
from tensorflow_datasets.scripts.cli import new

FLAGS = flags.FLAGS


@dataclasses.dataclass(frozen=True, kw_only=True)
class _DummyCommand:
  """Dummy command to avoid `command is MISSING` error."""

  pass


version_field = simple_parsing.field(
    action='version',
    version='TensorFlow Datasets: ' + tfds.__version__,
    help='The version of the TensorFlow Datasets package.',
)


@dataclasses.dataclass(frozen=True, kw_only=True)
class Args(cli_utils.Args):
  """Tensorflow Datasets CLI tool."""

  version: str = version_field
  """The version of the TensorFlow Datasets package."""

  dry_run: bool = simple_parsing.flag(default=False)
  """If True, print the parsed arguments and exit."""

  command: build.Args | new.Args | convert_format.Args | croissant.CmdArgs = (
      simple_parsing.subparsers(
          {
              'build': build.Args,
              'new': new.Args,
              'convert_format': convert_format.Args,
              'build_croissant': croissant.CmdArgs,
          },
          default_factory=_DummyCommand,
      )
  )
  """The command to execute."""

  def execute(self) -> None:
    """Run the command."""
    if self.dry_run:
      print(self)
    # When no command is given, print the help message.
    elif isinstance(self.command, _DummyCommand):
      _parse_flags(['', '--help'])
    else:
      self.command.execute()


_parse_flags = cli_utils.make_flags_parser(
    Args, description='Tensorflow Datasets CLI tool'
)


def main(args: Args) -> None:

  # From the CLI, all datasets are visible
  tfds.core.visibility.set_availables([
      tfds.core.visibility.DatasetType.TFDS_PUBLIC,
      tfds.core.visibility.DatasetType.COMMUNITY_PUBLIC,
  ])

  # By default, ABSL won't display any `logging.info` unless the
  # user explicitly set `--logtostderr`.
  # For usability, we activate by default the python log information, but
  # not the C++ ones (which are too verbose).
  # `logtostderr` may not be defined if `main()` is called directly without
  # `absl.run` (e.g. open source `pytest` tests)
  if not FLAGS.is_parsed() or (
      # If user explicitly request logs, keep C++ logger
      not FLAGS.logtostderr
      and not FLAGS.alsologtostderr
  ):
    # Using cleaner, less verbose logger
    formatter = python_logging.Formatter(
        '{levelname}[{filename}]: {message}', style='{'
    )
    logging.use_python_logging(quiet=True)
    logging.set_verbosity(logging.INFO)
    python_handler = logging.get_absl_handler().python_handler
    python_handler.setFormatter(formatter)
    # Replace `sys.stderr` by the TQDM file
    new_stream = tfds.core.utils.tqdm_utils.TqdmStream()
    python_handler.setStream(new_stream)

  args.execute()


def launch_cli() -> None:
  """Parse arguments and launch the CLI main function."""
  app.run(main, flags_parser=_parse_flags)


if __name__ == '__main__':
  # Entry-points in `setup.py` launch the `launch_cli()` function directly, so
  # the code in `if __name__ == '__main__'` is never executed.
  launch_cli()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imagenet val. annotated by ReaL labels (https://arxiv.org/abs/2006.07159)."""

import json
import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds


_REAL_LABELS_URL = 'https://raw.githubusercontent.com/google-research/reassessed-imagenet/master/real.json'


class Builder(tfds.core.GeneratorBasedBuilder):
  """ImageNet validation images with ReaL labels."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release',
  }

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  manual_dir should contain `ILSVRC2012_img_val.tar` file.
  You need to register on http://www.image-net.org/download-images in order
  to get the link to download the dataset.
  """

  def _info(self):
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'original_label': tfds.features.ClassLabel(names_file=names_file),
            'real_label': tfds.features.Sequence(
                tfds.features.ClassLabel(names_file=names_file)
            ),
            'file_name': tfds.features.Text(),
        }),
        supervised_keys=('image', 'real_label'),
        homepage='https://github.com/google-research/reassessed-imagenet',
    )

  def _get_real_labels(self, dl_manager):
    with tf.io.gfile.GFile(dl_manager.download(_REAL_LABELS_URL), 'r') as f:
      # ReaL labels are ordered in the lexicographical order.
      return {
          'ILSVRC2012_val_{:08}.JPEG'.format(i + 1): labels
          for i, labels in enumerate(json.load(f))
      }

  def _split_generators(self, dl_manager):
    val_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_val.tar')
    if not tf.io.gfile.exists(val_path):
      raise AssertionError(
          'ImageNet requires manual download of the data. Please download '
          'the train and val set and place them into: {}'.format(val_path)
      )
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'archive': dl_manager.iter_archive(val_path),
                'original_labels': imagenet_common.get_validation_labels(
                    val_path
                ),
                'real_labels': self._get_real_labels(dl_manager),
            },
        ),
    ]

  def _generate_examples(self, archive, original_labels, real_labels):
    for fname, fobj in archive:
      record = {
          'file_name': fname,
          'image': fobj,
          'original_label': original_labels[fname],
          'real_label': real_labels[fname],
      }
      yield fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Version info (development version).

All users importing TFDS from `tfds-nightly` or synced to head will see
the `-nightly` suffix.

This file is replaced by `version_stable.py` for stable releases
(`tensorflow-datasets`) on PyPI.
"""

# We follow Semantic Versioning (https://semver.org/spec/v2.0.0.html)
_MAJOR_VERSION = '4'
_MINOR_VERSION = '9'
_PATCH_VERSION = '9'

__version__ = (
    '.'.join([
        _MAJOR_VERSION,
        _MINOR_VERSION,
        _PATCH_VERSION,
    ])
    + '+nightly'
)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NEWSROOM Dataset."""

import json
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DOCUMENT = "text"
_SUMMARY = "summary"
_ADDITIONAL_TEXT_FEATURES = [
    "title",
    "url",
    "date",
    "density_bin",
    "coverage_bin",
    "compression_bin",
]
_ADDITIONAL_FLOAT_FEATURES = [
    "density",
    "coverage",
    "compression",
]


class Builder(tfds.core.GeneratorBasedBuilder):
  """NEWSROOM Dataset."""

  VERSION = tfds.core.Version("1.0.0")
  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  You should download the dataset from https://summari.es/download/
  The webpage requires registration.
  After downloading, please put dev.jsonl, test.jsonl and train.jsonl
  files in the manual_dir.
  """

  def _info(self):
    features = {
        k: tfds.features.Text()
        for k in [_DOCUMENT, _SUMMARY] + _ADDITIONAL_TEXT_FEATURES
    }
    features.update(
        {
            k: tfds.features.Tensor(shape=[], dtype=np.float32)
            for k in _ADDITIONAL_FLOAT_FEATURES
        }
    )
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(features),
        supervised_keys=(_DOCUMENT, _SUMMARY),
        homepage="https://summari.es",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                "input_file": os.path.join(dl_manager.manual_dir, "train.jsonl")
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                "input_file": os.path.join(dl_manager.manual_dir, "dev.jsonl")
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "input_file": os.path.join(dl_manager.manual_dir, "test.jsonl")
            },
        ),
    ]

  def _generate_examples(self, input_file=None):
    """Yields examples."""
    with epath.Path(input_file).open() as f:
      for i, line in enumerate(f):
        d = json.loads(line)
        # fields are "url", "archive", "title", "date", "text",
        #  "compression_bin", "density_bin", "summary", "density",
        #  "compression', "coverage", "coverage_bin",
        yield i, {
            k: d[k]
            for k in [_DOCUMENT, _SUMMARY]
            + _ADDITIONAL_TEXT_FEATURES
            + _ADDITIONAL_FLOAT_FEATURES
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""kdd_cup_99 dataset."""

import csv
import gzip

import numpy as np
from tensorflow_datasets.core.utils import bool_utils
import tensorflow_datasets.public_api as tfds

_TRAIN_URL = 'http://kdd.ics.uci.edu/databases/kddcup99/kddcup.data.gz'
_TEST_URL = 'http://kdd.ics.uci.edu/databases/kddcup99/corrected.gz'


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for kdd_cup_99 dataset."""

  VERSION = tfds.core.Version('1.0.1')
  RELEASE_NOTES = {
      '1.0.1': (
          'Fixes parsing of boolean fields `land`, `logged_in`, '
          '`root_shell`, `is_hot_login` and `is_guest_login`.'
      ),
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'duration': np.int32,
            'protocol_type': tfds.features.ClassLabel(
                names=['icmp', 'tcp', 'udp']
            ),
            'service': tfds.features.ClassLabel(
                names=[
                    'aol',
                    'auth',
                    'bgp',
                    'courier',
                    'csnet_ns',
                    'ctf',
                    'daytime',
                    'discard',
                    'domain',
                    'domain_u',
                    'echo',
                    'eco_i',
                    'ecr_i',
                    'efs',
                    'exec',
                    'finger',
                    'ftp',
                    'ftp_data',
                    'gopher',
                    'harvest',
                    'hostnames',
                    'http',
                    'http_2784',
                    'http_443',
                    'http_8001',
                    'icmp',
                    'imap4',
                    'IRC',
                    'iso_tsap',
                    'klogin',
                    'kshell',
                    'ldap',
                    'link',
                    'login',
                    'mtp',
                    'name',
                    'netbios_dgm',
                    'netbios_ns',
                    'netbios_ssn',
                    'netstat',
                    'nnsp',
                    'nntp',
                    'ntp_u',
                    'other',
                    'pm_dump',
                    'pop_2',
                    'pop_3',
                    'printer',
                    'private',
                    'red_i',
                    'remote_job',
                    'rje',
                    'shell',
                    'smtp',
                    'sql_net',
                    'ssh',
                    'sunrpc',
                    'supdup',
                    'systat',
                    'telnet',
                    'tftp_u',
                    'time',
                    'tim_i',
                    'urh_i',
                    'urp_i',
                    'uucp',
                    'uucp_path',
                    'vmnet',
                    'whois',
                    'X11',
                    'Z39_50',
                ]
            ),
            'flag': tfds.features.ClassLabel(
                names=[
                    'OTH',
                    'REJ',
                    'RSTO',
                    'RSTOS0',
                    'RSTR',
                    'S0',
                    'S1',
                    'S2',
                    'S3',
                    'SF',
                    'SH',
                ]
            ),
            'src_bytes': np.int32,
            'dst_bytes': np.int32,
            'land': np.bool_,
            'wrong_fragment': np.int32,
            'urgent': np.int32,
            'hot': np.int32,
            'num_failed_logins': np.int32,
            'logged_in': np.bool_,
            'num_compromised': np.int32,
            'root_shell': np.bool_,
            'su_attempted': np.int32,
            'num_root': np.int32,
            'num_file_creations': np.int32,
            'num_shells': np.int32,
            'num_access_files': np.int32,
            'num_outbound_cmds': np.int32,
            'is_hot_login': np.bool_,
            'is_guest_login': np.bool_,
            'count': np.int32,
            'srv_count': np.int32,
            'serror_rate': np.float32,
            'srv_serror_rate': np.float32,
            'rerror_rate': np.float32,
            'srv_rerror_rate': np.float32,
            'same_srv_rate': np.float32,
            'diff_srv_rate': np.float32,
            'srv_diff_host_rate': np.float32,
            'dst_host_count': np.int32,
            'dst_host_srv_count': np.int32,
            'dst_host_same_srv_rate': np.float32,
            'dst_host_diff_srv_rate': np.float32,
            'dst_host_same_src_port_rate': np.float32,
            'dst_host_srv_diff_host_rate': np.float32,
            'dst_host_serror_rate': np.float32,
            'dst_host_srv_serror_rate': np.float32,
            'dst_host_rerror_rate': np.float32,
            'dst_host_srv_rerror_rate': np.float32,
            'label': tfds.features.ClassLabel(
                names=[
                    'apache2',
                    'back',
                    'buffer_overflow',
                    'ftp_write',
                    'guess_passwd',
                    'httptunnel',
                    'imap',
                    'ipsweep',
                    'land',
                    'loadmodule',
                    'mailbomb',
                    'mscan',
                    'multihop',
                    'named',
                    'neptune',
                    'nmap',
                    'normal',
                    'perl',
                    'phf',
                    'pod',
                    'portsweep',
                    'processtable',
                    'ps',
                    'rootkit',
                    'saint',
                    'satan',
                    'sendmail',
                    'smurf',
                    'snmpgetattack',
                    'snmpguess',
                    'spy',
                    'sqlattack',
                    'teardrop',
                    'udpstorm',
                    'warezclient',
                    'warezmaster',
                    'worm',
                    'xlock',
                    'xsnoop',
                    'xterm',
                ]
            ),
        }),
        supervised_keys=None,  # Set to `None` to disable
        homepage='https://kdd.ics.uci.edu/databases/kddcup99/kddcup99.html',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    paths = dl_manager.download({
        'train': _TRAIN_URL,
        'test': _TEST_URL,
    })

    return {
        tfds.Split.TRAIN: self._generate_examples(paths['train']),
        tfds.Split.TEST: self._generate_examples(paths['test']),
    }

  def _generate_examples(self, gz_path):
    """Yields examples."""
    with gz_path.open('rb') as f:
      with gzip.open(f, 'rt', newline='') as gz:
        reader = csv.DictReader(gz, self.info.features)
        for index, row in enumerate(reader):
          row['land'] = bool_utils.parse_bool(row['land'])
          row['logged_in'] = bool_utils.parse_bool(row['logged_in'])
          row['root_shell'] = bool_utils.parse_bool(row['root_shell'])
          row['is_hot_login'] = bool_utils.parse_bool(row['is_hot_login'])
          row['is_guest_login'] = bool_utils.parse_bool(row['is_guest_login'])
          row['label'] = str(row['label']).rstrip('.')
          yield index, row



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""race dataset."""

import json
import tensorflow_datasets.public_api as tfds

_MODULES = [
    "high",
    "middle",
]


def _make_builder_config(module):
  return tfds.core.BuilderConfig(
      name=module,
      description="Builder config for RACE dataset.",
  )


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for race dataset."""

  VERSION = tfds.core.Version("2.0.0")
  RELEASE_NOTES = {
      "2.0.0": "Add the example id.",
      "1.0.0": "Initial release.",
  }
  BUILDER_CONFIGS = [_make_builder_config(module) for module in _MODULES]

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "article": tfds.features.Text(),
            "questions": tfds.features.Sequence(tfds.features.Text()),
            "answers": tfds.features.Sequence(tfds.features.Text()),
            "options": tfds.features.Sequence(
                tfds.features.Sequence(tfds.features.Text())
            ),
            "example_id": tfds.features.Text(),
        }),
        supervised_keys=None,  # Set to `None` to disable
        homepage="https://www.cs.cmu.edu/~glai1/data/race/",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        "http://www.cs.cmu.edu/~glai1/data/race/RACE.tar.gz"
    )

    path = path / "RACE"
    task = self.builder_config.name

    return {
        "train": self._generate_examples(path / "train" / task),
        "dev": self._generate_examples(path / "dev" / task),
        "test": self._generate_examples(path / "test" / task),
    }

  def _generate_examples(self, path):
    """Yields examples."""

    for file in path.iterdir():
      # Each file is one example and only has one line of the content.
      row = json.loads(file.read_text())
      yield row["id"], {
          "article": row["article"],
          "questions": row["questions"],
          "answers": row["answers"],
          "options": row["options"],
          "example_id": row["id"],
      }
